# -*- coding: utf-8 -*-

import sys

from gerritaction.action.action import Action, ActionException
from gerritaction.cmd.argument import Argument
from gerritaction.cmd.banner import BANNER
from gerritaction.config.config import Config, ConfigException
from gerritaction.logger.logger import Logger


def main():
    print(BANNER)

    argument = Argument()
    arg = argument.parse(sys.argv)

    try:
        config = Config()
        config.config_file = arg.config_file
        config.gerrit_action = arg.gerrit_action
        config.gerrit_query = arg.gerrit_query
    except ConfigException as e:
        Logger.error(str(e))
        return -1

    Logger.info("action running")

    try:
        action = Action(config)
        action.run()
    except ActionException as e:
        Logger.error(str(e))
        return -2

    Logger.info("action exiting")

    return 0



# -*- coding: utf-8 -*-

import argparse

from gerritaction.__version__ import __version__


class Argument(object):
    def __init__(self):
        self._parser = argparse.ArgumentParser(description="Gerrit Action")
        self._add()

    def _add(self):
        self._parser.add_argument(
            "--config-file",
            action="store",
            dest="config_file",
            help="config file (.yml)",
            required=True,
        )
        self._parser.add_argument(
            "--gerrit-action",
            action="store",
            dest="gerrit_action",
            help="gerrit action (add-reviewer:account-id,... delete-reviewer:account-id,... add-attention:account-id,... remove-attention:account-id,... approve-change:Code-Review=+2,... submit-change)",
            required=True,
        )
        self._parser.add_argument(
            "--gerrit-query",
            action="store",
            dest="gerrit_query",
            help="gerrit query (status:open since:2021-01-01 until:2021-01-02)",
            required=True,
        )
        self._parser.add_argument(
            "-v", "--version", action="version", version=__version__
        )

    def parse(self, argv):
        return self._parser.parse_args(argv[1:])



# -*- coding: utf-8 -*-


class Proto:
    ADD_REVIEWER = "add-reviewer"
    DELETE_REVIEWER = "delete-reviewer"
    ADD_ATTENTION = "add-attention"
    REMOVE_ATTENTION = "remove-attention"
    APPROVE_CHANGE = "approve-change"
    SUBMIT_CHANGE = "submit-change"


class Separator:
    ACTION = ":"
    CONTENT = ","
    GROUP = " "
    LABEL = "="



# -*- coding: utf-8 -*-

import os
import yaml


class ConfigFile:
    APIVERSION = "apiVersion"
    KIND = "kind"
    METADATA = "metadata"
    SPEC = "spec"


class MetaData:
    NAME = "name"


class ConfigException(Exception):
    def __init__(self, info):
        super().__init__(self)
        self._info = info

    def __str__(self):
        return self._info


class Config(object):
    def __init__(self):
        self._config_file = None
        self._gerrit_action = None
        self._gerrit_query = None

    @property
    def config_file(self):
        return self._config_file

    @config_file.setter
    def config_file(self, name):
        if not isinstance(name, str) or len(name.strip()) == 0:
            raise ConfigException("config invalid")
        name = name.strip()
        if not name.endswith(".yml") and not name.endswith(".yaml"):
            raise ConfigException("suffix invalid")
        if not os.path.exists(name):
            raise ConfigException("%s not found" % name)
        with open(name) as file:
            self._config_file = yaml.load(file, Loader=yaml.FullLoader)
        if self._config_file is None:
            raise ConfigException("config invalid")

    @property
    def gerrit_action(self):
        return self._gerrit_action

    @gerrit_action.setter
    def gerrit_action(self, data):
        if not isinstance(data, str) or len(data.strip()) == 0:
            raise ConfigException("action invalid")
        self._gerrit_action = data.strip()

    @property
    def gerrit_query(self):
        return self._gerrit_query

    @gerrit_query.setter
    def gerrit_query(self, data):
        if not isinstance(data, str) or len(data.strip()) == 0:
            raise ConfigException("query invalid")
        self._gerrit_query = data.strip()



# -*- coding: utf-8 -*-

from ..gerrit.gerrit import Gerrit
from ..proto.proto import Proto, Separator

from gerritaction.config.config import ConfigFile


class ActionException(Exception):
    def __init__(self, info):
        super().__init__(self)
        self._info = info

    def __str__(self):
        return self._info


class Action(object):
    def __init__(self, config=None):
        if config is None or config.config_file.get(ConfigFile.SPEC, None) is None:
            raise ActionException("config invalid")
        self._config = config
        self._gerrit = Gerrit(self._config.config_file[ConfigFile.SPEC])
        self._changes = self._query_changes()
        if self._changes is None:
            raise ActionException("changes invalid")

    def _query_changes(self):
        return self._gerrit.query_changes(search=self._config.gerrit_query, start=0)

    def _add_reviewer(self, accounts):
        for account in accounts:
            for change in self._changes:
                _ = self._gerrit.add_reviewer(change, account)

    def _delete_reviewer(self, accounts):
        for account in accounts:
            for change in self._changes:
                _ = self._gerrit.delete_reviewer(change, account)

    def _add_attention(self, accounts):
        for account in accounts:
            for change in self._changes:
                _ = self._gerrit.add_attention(change, account)

    def _remove_attention(self, accounts):
        for account in accounts:
            for change in self._changes:
                _ = self._gerrit.remove_attention(change, account)

    def _approve_change(self, labels):
        for change in self._changes:
            _ = self._gerrit.approve_change(change, labels)

    def _submit_change(self):
        for change in self._changes:
            _ = self._gerrit.submit_change(change)

    def run(self):
        for item in self._config.gerrit_action.split(Separator.GROUP):
            if len(item.split(Separator.ACTION)) > 2:
                raise ActionException("action invalid")
            if Proto.SUBMIT_CHANGE in item:
                self._submit_change()
                continue
            action, content = item.split(Separator.ACTION)
            if action == Proto.ADD_REVIEWER:
                self._add_reviewer(content.split(Separator.CONTENT))
            elif action == Proto.DELETE_REVIEWER:
                self._delete_reviewer(content.split(Separator.CONTENT))
            elif action == Proto.ADD_ATTENTION:
                self._add_attention(content.split(Separator.CONTENT))
            elif action == Proto.REMOVE_ATTENTION:
                self._remove_attention(content.split(Separator.CONTENT))
            elif action == Proto.APPROVE_CHANGE:
                self._approve_change(content.split(Separator.CONTENT))
            else:
                raise ActionException("action invalid")



# -*- coding: utf-8 -*-

import json
import requests

from ..proto.proto import Separator

from gerritaction.logger.logger import Logger


class GerritException(Exception):
    def __init__(self, info):
        super().__init__(self)
        self._info = info

    def __str__(self):
        return self._info


class Gerrit(object):
    _QUERY_LIMIT = 1000

    def __init__(self, config):
        if config is None or config.get("gerrit", None) is None:
            raise GerritException("config invalid")
        self._config = config["gerrit"]
        self._host = str(self._config.get("host", "http://127.0.0.1"))
        self._port = str(self._config.get("port", 8080))
        self._user = str(self._config.get("user", ""))
        self._pass = str(self._config.get("pass", ""))
        self._query = self._config.get("query", {"option": ["CURRENT_REVISION"]})
        if len(self._pass) != 0 and len(self._user) != 0:
            self._url = self._host + ":" + self._port + "/a"
        else:
            self._url = self._host + ":" + self._port

    def query_changes(self, search, start):
        def _helper(search, start):
            payload = {
                "o": self._query["option"],
                "q": search,
                "start": start,
                "n": self._QUERY_LIMIT,
            }
            if len(self._pass) != 0 and len(self._user) != 0:
                response = requests.get(
                    url=self._url + "/changes/",
                    auth=(self._user, self._pass),
                    params=payload,
                )
            else:
                response = requests.get(url=self._url + "/changes/", params=payload)
            if response.status_code != requests.codes.ok:
                Logger.error("failed to query change with search %s" % search)
                return None
            return json.loads(response.text.replace(")]}'", ""))

        buf = _helper(search, start)
        if buf is None or len(buf) == 0:
            return []
        if buf[-1].get("_more_changes", False) is False:
            return buf
        buf.extend(self.query_changes(search, start + len(buf)))
        return buf

    def get_detail(self, change):
        if len(self._pass) != 0 and len(self._user) != 0:
            response = requests.get(
                url=self._url + "/changes/" + str(change["_number"]) + "/detail",
                auth=(self._user, self._pass),
            )
        else:
            response = requests.get(
                url=self._url + "/changes/" + str(change["_number"]) + "/detail"
            )
        if response.status_code != requests.codes.ok:
            Logger.error("failed to get detail for change %s" % change["_number"])
            return None
        return json.loads(response.text.replace(")]}'", ""))

    def add_reviewer(self, change, account):
        args = {"reviewer": account}
        if len(self._pass) != 0 and len(self._user) != 0:
            response = requests.post(
                url=self._url + "/changes/" + str(change["_number"]) + "/reviewers",
                auth=(self._user, self._pass),
                json=args,
            )
        else:
            response = requests.post(
                url=self._url + "/changes/" + str(change["_number"]) + "/reviewers",
                json=args,
            )
        if response.status_code != requests.codes.ok:
            Logger.error(
                "failed to add reviewer %s to change %s" % (account, change["_number"])
            )
            return None
        return json.loads(response.text.replace(")]}'", ""))

    def delete_reviewer(self, change, account):
        if len(self._pass) != 0 and len(self._user) != 0:
            response = requests.delete(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/reviewers/"
                + str(account),
                auth=(self._user, self._pass),
            )
        else:
            response = requests.delete(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/reviewers/"
                + str(account)
            )
        if (
            response.status_code != requests.codes.ok
            and response.status_code != requests.codes.no_content
        ):
            Logger.error(
                "failed to delete reviewer %s from change %s"
                % (account, change["_number"])
            )
            return None
        return json.loads("{}")

    def add_attention(self, change, account):
        args = {"user": account, "reason": "attention"}
        if len(self._pass) != 0 and len(self._user) != 0:
            response = requests.post(
                url=self._url + "/changes/" + str(change["_number"]) + "/attention",
                auth=(self._user, self._pass),
                json=args,
            )
        else:
            response = requests.post(
                url=self._url + "/changes/" + str(change["_number"]) + "/attention",
                json=args,
            )
        if response.status_code != requests.codes.ok:
            Logger.error(
                "failed to add attention %s to change %s" % (account, change["_number"])
            )
            return None
        return json.loads(response.text.replace(")]}'", ""))

    def remove_attention(self, change, account):
        args = {"reason": "ignore"}
        if len(self._pass) != 0 and len(self._user) != 0:
            response = requests.delete(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/attention/"
                + str(account),
                auth=(self._user, self._pass),
                json=args,
            )
        else:
            response = requests.delete(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/attention/"
                + str(account),
                json=args,
            )
        if (
            response.status_code != requests.codes.ok
            and response.status_code != requests.codes.no_content
        ):
            Logger.error(
                "failed to remove attention %s from change %s"
                % (account, change["_number"])
            )
            return None
        return json.loads("{}")

    def approve_change(self, change, labels):
        def _helper(labels):
            buf = {}
            for item in labels:
                key, val = item.split(Separator.LABEL)
                buf[key] = int(val)
            return buf

        args = {
            "comments": {},
            "labels": _helper(labels),
            "message": "Approved by %s" % self._user,
            "tag": "",
        }
        if len(self._pass) != 0 and len(self._user) != 0:
            response = requests.post(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/revisions/"
                + str(change["current_revision"])
                + "/review",
                auth=(self._user, self._pass),
                json=args,
            )
        else:
            response = requests.post(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/revisions/"
                + str(change["current_revision"])
                + "/review",
                json=args,
            )
        if response.status_code != requests.codes.ok:
            Logger.error(
                "failed to approve change %s by account %s"
                % (change["_number"], self._user)
            )
            return None
        return json.loads(response.text.replace(")]}'", ""))

    def submit_change(self, change):
        if len(self._pass) != 0 and len(self._user) != 0:
            response = requests.post(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/revisions/"
                + str(change["current_revision"])
                + "/submit",
                auth=(self._user, self._pass),
                json=None,
            )
        else:
            response = requests.post(
                url=self._url
                + "/changes/"
                + str(change["_number"])
                + "/revisions/"
                + str(change["current_revision"])
                + "/submit",
                json=None,
            )
        if response.status_code != requests.codes.ok:
            Logger.error(
                "failed to submit change %s by account %s"
                % (change["_number"], self._user)
            )
            return None
        return json.loads(response.text.replace(")]}'", ""))
